from pathlib import Path
from unittest import mock

import os
import pycarol


def _check_constants(carol: pycarol.Carol) -> None:
    assert carol._current_user is None
    assert carol._tenant is None
    assert carol._user_agent == f"pyCarol/{pycarol.__version__}"
    assert carol.org is None
    assert carol.response is None
    assert carol.session is None


@mock.patch("pycarol.carol._set_host")
@mock.patch("pycarol.carol._prepare_auth")
def test_init_carol_envvar(mock_prepareauth, mock_sethost) -> None:

    tmp_filepath = "/tmp/.env_test"
    with open(tmp_filepath, "w", encoding="utf-8") as file:
        file.write("CAROLAPPNAME=APPNAME\n")
        file.write("CAROL_DOMAIN=ENV\n")
        file.write("CAROLORGANIZATION=ORG\n")
        file.write("CAROLTENANT=TENANT\n")

    carol = pycarol.Carol(dotenv_path=tmp_filepath)

    Path(tmp_filepath).unlink()

    _check_constants(carol)
    assert carol._host_string is None
    assert carol._is_org_level is False
    assert carol.app_name == "APPNAME"
    assert carol.auth == mock_prepareauth.return_value
    assert carol.connector_id == mock_prepareauth.return_value.connector_id
    assert carol.domain == "TENANT"
    assert carol.environment == "ENV"
    assert carol.host == mock_sethost.return_value
    assert carol.organization == "ORG"
    assert carol.port == 443
    assert carol.verbose is False

    del os.environ["CAROLAPPNAME"]
    del os.environ["CAROL_DOMAIN"]
    del os.environ["CAROLORGANIZATION"]
    del os.environ["CAROLTENANT"]


@mock.patch("pycarol.carol._prepare_auth")
@mock.patch("pycarol.carol._set_host")
def test_init_carol_args(mock_set_host, mock_getauth) -> None:
    mock_auth = mock.MagicMock()
    carol = pycarol.Carol(
        "domain",
        "app_name",
        mock_auth,
        "connector_id",
        port=444,
        verbose=True,
        organization="organization",
        environment="environment",
        host="host",
        password="password",
        api_key="api_key",
        org_level=False,
    )

    _check_constants(carol)
    assert carol._host_string == "host"
    assert carol._is_org_level is False
    assert carol.app_name == "app_name"
    assert carol.auth == mock_getauth.return_value
    assert carol.connector_id == mock_getauth.return_value.connector_id
    assert carol.domain == "domain"
    assert carol.environment == "environment"
    assert carol.host == mock_set_host.return_value
    assert carol.organization == "organization"
    assert carol.port == 444
    assert carol.verbose is True


@mock.patch("pycarol.carol._prepare_auth")
@mock.patch("pycarol.carol._set_host")
def test_init_carol_exceptions(mock_set_host, mock_getauth) -> None:
    try:
        pycarol.Carol()
        assert "Should have thrown exception" == ""
    except Exception as exception:
        assert isinstance(exception, pycarol.exceptions.MissingInfoCarolException)


def test_set_host() -> None:
    ret = pycarol.carol._set_host("environment", "domain", "host", "org")
    assert ret == "host"
    ret = pycarol.carol._set_host("environment", "domain", None, "org")
    assert ret == "org.environment"
    ret = pycarol.carol._set_host("environment", "domain", None, None)
    assert ret == "domain.environment"


@mock.patch("pycarol.carol.ApiKeyAuth")
@mock.patch("pycarol.carol.PwdAuth")
def test_set_auth(mock_pwdauth, mock_apikeyauth) -> None:
    try:
        pycarol.carol._prepare_auth()
        assert "Should have thrown exception" == ""
    except Exception as exception:
        assert isinstance(exception, pycarol.exceptions.MissingInfoCarolException)

    auth = mock.MagicMock()
    ret = pycarol.carol._prepare_auth(auth=auth)
    assert ret == auth

    ret = pycarol.carol._prepare_auth(app_oauth="oauth")
    assert ret == mock_apikeyauth.return_value

    ret = pycarol.carol._prepare_auth(username="user", password="pass")
    assert ret == mock_pwdauth.return_value


def test_get_current() -> None:
    carol = mock.MagicMock()
    carol._current_env = lambda: {"mdmName": "name1", "mdmId": "id1"}
    carol._current_org = lambda: {"mdmName": "name2", "mdmId": "id2"}
    try:
        pycarol.carol.Carol.get_current(carol, "coco")
        assert "Should have thrown exception" == ""
    except Exception as exception:
        assert isinstance(exception, ValueError)

    ret = pycarol.carol.Carol.get_current(carol, "env")
    assert ret["env_name"] == "name1"
    assert ret["env_id"] == "id1"
    assert ret["org_name"] is None
    assert ret["org_id"] is None

    ret = pycarol.carol.Carol.get_current(carol, "all")
    assert ret["env_name"] == "name1"
    assert ret["env_id"] == "id1"
    assert ret["org_name"] == "name2"
    assert ret["org_id"] == "id2"


@mock.patch("pycarol.carol._set_host")
@mock.patch("pycarol.carol.Organization")
@mock.patch("pycarol.carol.Tenant")
@mock.patch("pycarol.carol.Carol.get_current")
def test_switch_env(mock_getcurrent, mock_tenant, mock_org, mock_sethost) -> None:
    carol = mock.MagicMock()
    carol._current_org = lambda: {"mdmName": "name"}
    carol.org = None
    carol_ = pycarol.carol.Carol.switch_environment(carol)
    assert carol_.org == mock_org.return_value.get_organization_info.return_value
    assert carol_.app_name is None
    assert carol_.domain is None
    assert carol_.host == mock_sethost.return_value
    assert carol_.organization == "name"
    assert carol_._tenant is None
    assert carol_._is_org_level is True

    carol2 = mock.MagicMock()
    carol2._current_org = lambda: {"mdmName": "name"}
    carol2.org = None
    carol_ = pycarol.carol.Carol.switch_environment(
        carol2, org_id="org", env_name="name"
    )
    assert carol_.org == mock_org.return_value.get_organization_info.return_value
    assert carol_.app_name is None
    assert carol_.domain == "name"
    assert carol_.host == mock_sethost.return_value
    assert carol_.organization == "name"
    assert carol_._tenant == carol2._current_env.return_value
    assert carol_._is_org_level is False



"""Contain all the classes to query data from RT layer in Carol."""
import copy
from datetime import datetime
import json
import itertools
import typing as T

from retry import retry

from .connectors import Connectors
from .exceptions import NoScrollIdException, RepeatedMDMIdsException
from .filter import Filter, MAXIMUM, MINIMUM, TYPE_FILTER, TERM_FILTER
from .filter import RANGE_FILTER as RF
from .named_query import NamedQuery
from .utils.miscellaneous import ranges


def delete_golden(carol, dm_name, now=None):
    """
    Delete Golden records.

    It Will delete all golden records of a given data model based on lastUpdate.

    Args:

        carol: `pycarol.carol.Carol`
            Carol instance
        dm_name: `str`
            Data model name
        now: `str`
            Delete records where last update is less the `now`. Any date time ISO format is accepted.

    Usage:

    .. code:: python

        from pycarol.query import delete_golden
        from pycarol.auth.PwdAuth import PwdAuth
        from pycarol.carol import Carol
        login = Carol()
        delete_golden(login, dm_name=my_dm)

        #To delete based on a date.
        delete_golden(login, dm_name=my_dm, now='2018-11-16')

    Attention:

        This API will delete all data in the DataModel, and if there is a DataModel View related to this DataModel
        one needs to reprocess it.

    """

    if now is None:
        now = datetime.utcnow().isoformat(timespec="seconds")

    json_query = (
        Filter.Builder()
        .should(TYPE_FILTER(value=dm_name + "Golden"))
        .should(TYPE_FILTER(value=dm_name + "Master"))
        .must(RF("mdmLastUpdated", [None, now]))
        .build()
        .to_json()
    )

    try:
        Query(carol).delete(json_query)
    except:
        # if it it too many records, one would have a timeout but the records will be deleted anyway
        pass

    json_query = (
        Filter.Builder()
        .type(dm_name + "Rejected")
        .must(RF("mdmLastUpdated", [None, now]))
        .build()
        .to_json()
    )

    try:
        Query(carol, index_type="STAGING").delete(json_query)
    except:
        # if it it too many records, one would have a timeout but the records will be deleted anyway
        pass


class Query:
    """
    Class to query data from Carol.

    This class can be used to query data from data models and stagings tables, query using named queries and delete
    records.

    Args:

        carol: carol: Carol object
            Carol object.
        max_hits:  `int`, default float('inf')
            Number of records that will be downloaded.
        offset: `int`, default 0
            Offset for pagination. Only used when `scrollable=False`
        page_size: `int`, default 100
            Number of records downloaded in each pagination. The maximum value is 1000
        sort_order: `str`, default 'ASC'
            Sort ascending ('ASC') vs. descending ('DESC').
        sort_by: `str`,  default `None`
            Name to sort by.
        scrollable: `bool`, default True
            Use scroll for pagination. This should be the main way of doing, unless you are querying few data.
        index_type: `str`, default 'MASTER'
            Query data from 'MASTER', 'STAGING'
        only_hits: `bool`, default 'True'
            Return only results in the response path $hits.mdmGoldenFieldAndValues
        fields: `list`, default `None`
            Fields to return in response. e.g., ["mdmGoldenFieldAndValues.mdmtaxid", "mdmGoldenFieldAndValues.date"]
        get_aggs: `bool`, default `False`
            To be used if the query/named query has aggravations
        save_results: `bool`, default `False`
            If save the result of the query in the file specified in `filename`
        filename: `str`, default `query_result.json`
            File path to save the response.
        print_status: `bool`, default `True`
            Print the number of records in each interaction.
        safe_check: `bool`, default `False`
            To be used if there are repeated records (same mdmId)
        get_errors: `bool`, default `False`
            To get the errors in the goldenRecords, if any.
        flush_result: `bool`, default `False`
            To be used with save_results, it will not copy the result to memory, only to the file.
        use_stream: `bool`, default `False`
            Use the stram of data.
        get_times: `bool`, default `False`
            It will create a list of times that each pagination took.
        kwargs: `dict`
            Extra parameters to be passed to Carol.call_api

    """

    def __init__(
        self,
        carol,
        max_hits=float("inf"),
        offset=0,
        page_size=100,
        sort_order="ASC",
        sort_by=None,
        scrollable=True,
        index_type="MASTER",
        only_hits=True,
        fields=None,
        get_aggs=False,
        save_results=False,
        filename="query_result.json",
        print_status=True,
        safe_check=False,
        get_errors=False,
        flush_result=False,
        use_stream=False,
        get_times=False,
        **kwargs,
    ):

        self.carol = carol
        self.max_hits = max_hits
        self.offset = offset
        self.page_size = page_size
        self.sort_order = sort_order
        self.sort_by = sort_by
        self.scrollable = scrollable
        self.index_type = index_type
        self.only_hits = only_hits
        self.fields = fields
        self.get_aggs = get_aggs
        self.use_stream = use_stream

        self.save_results = save_results
        self.filename = filename
        self.print_status = print_status
        self.safe_check = safe_check
        self.get_errors = get_errors
        self.flush_result = flush_result
        self.get_times = get_times
        self.named_query = None
        self.callback = None
        self.query_errors: T.Dict[str, T.Any] = {}

        # Crated to send to the Rest API
        self.query_params = None
        self.drop_list = None
        self.json_query = None
        self.total_hits = None
        self.query_times = []
        self.kwargs = kwargs

        self.results = []

        if self.max_hits == float("inf"):
            self.get_all = True
        else:
            self.get_all = False

    def _build_query_params(self):
        self.query_params = {
            "offset": self.offset,
            "pageSize": self.page_size,
            "sortOrder": self.sort_order,
            "indexType": self.index_type,
        }

        if self.sort_by is not None:
            self.query_params["sortBy"] = self.sort_by
        if self.scrollable:
            self.query_params["scrollable"] = self.scrollable
        if self.fields:
            self.query_params["fields"] = self.fields

    def _build_return_fields(self):
        if isinstance(self.fields, str):
            self.fields = [self.fields]

        if self.fields is not None:
            self.fields = ",".join(self.fields)

    def page(self, offset=0):
        """
        Get only one page of the result using offset.

        Args:
            offset: `int`, default 0
                Offset to get. To properly paginate manually, offset should be `offset + page_size`.

        Returns:
            Query json response
        """

        self.offset = offset
        self.scrollable = False

        self.results = []
        if self.json_query is None:
            raise ValueError("You must call query() or named() before calling page()")

        self._build_return_fields()
        self._build_query_params()

        if self.named_query is None:
            url_filter = "v2/queries/filter"
        else:
            url_filter = f"v2/queries/named/{self.named_query}"

        result = self.carol.call_api(
            url_filter,
            data=self.json_query,
            params=self.query_params,
            timeout=240,
            method_whitelist=frozenset(["POST"]),
            **self.kwargs,
        )

        if self.only_hits is True:
            result = result["hits"]
            return [
                elem.get("mdmGoldenFieldAndValues", elem)
                for elem in result
                if elem.get("mdmGoldenFieldAndValues", None)
            ]

        else:
            return result

    def go(self, callback: T.Optional[T.Callable] = None) -> "Query":
        """Run the query.

        Args:
            callback: This function will receive the current batch of records from the
                filter made.
        Returns: Query self.
        """
        self.results = []
        if self.json_query is None:
            raise ValueError(
                "You must call all() or query() or named() before calling go()"
            )

        self._build_return_fields()
        self._build_query_params()

        if not self.offset == 0:
            raise ValueError(
                "It is not possible to use offset when using scroll for pagination"
            )
        self._scrollable_query_handler(callback)

        return self

    def _scrollable_query_handler(
        self, callback: T.Optional[T.Callable] = None
    ) -> None:
        if self.named_query is None:
            url_filter = "v2/queries/filter"
        else:
            url_filter = f"v2/queries/named/{self.named_query}"

        count = self.offset
        downloaded = 0
        set_param = True
        to_get = float("inf")
        self.query_errors = {}
        self.mdmId_list = []
        while count < to_get:
            raise_error = self.get_aggs and not self.only_hits
            result = self._query_request(url_filter, raise_error)

            if set_param is True:
                self.total_hits = result["totalHits"]
                to_get = min(self.max_hits, self.total_hits)
                set_param = False

            count += result["count"]
            downloaded += result["count"]
            scroll_id = result.get("scrollId", None)
            url_filter = f"v2/queries/filter/{scroll_id}"

            if self.get_times is True:
                self.query_times.append(result.pop("took"))

            if self.safe_check is True:
                self.mdmId_list.extend([mdm_id["mdmId"] for mdm_id in result["hits"]])
                if len(self.mdmId_list) > len(set(self.mdmId_list)):
                    raise RepeatedMDMIdsException

            if self.get_errors is True:
                errors = {
                    elem.get("mdmId", elem): elem.get("mdmErrors", elem)
                    for elem in result["hits"]
                    if "mdmErrors" in elem and elem["mdmErrors"]
                }
                self.query_errors.update(errors)

            if self.only_hits is True:
                result = result["hits"]
                result = [
                    elem.get("mdmGoldenFieldAndValues", elem)
                    for elem in result
                    if elem.get("mdmGoldenFieldAndValues", None)
                ]  # get mdmGoldenFieldAndValues if not empty and if it exists

                if self.flush_result is False:
                    self.results.extend(result)
            else:
                result.pop("count")
                result.pop("totalHits")

                if self.flush_result is False:
                    self.results.append(result)

            if callback is not None:
                callback(result)

            if self.print_status is True:
                print(f"{downloaded}/{to_get}", end="\r")

            if self.save_results is True:
                _write_results(self.filename, result)

            if self.get_aggs is True and self.only_hits is False:
                break

            if scroll_id is None:
                print(result)
                raise NoScrollIdException

    @retry(exceptions=NoScrollIdException, tries=5)
    def _query_request(
        self, url: str, raise_error: bool
    ) -> T.Dict:
        result = self.carol.call_api(
            url,
            data=self.json_query,
            params=self.query_params,
            timeout=240,
            method_whitelist=frozenset(["POST"]),
            **self.kwargs,
        )

        scroll_id = result.get("scrollId", None)
        if raise_error is True and scroll_id is None:
            print(result)
            print("Retrying...")
            raise NoScrollIdException

        return result

    def check_total_hits(self, json_query, index_type="MASTER"):
        """
        Check the total hits for a given query
        :param json_query: Json object with the query to use
        :param index_type: Index type to query.
        :return: number of records for this query
        """

        querystring = {"indexType": index_type}
        self.json_query = json_query
        url_filter = "v2/queries/count"
        result = self.carol.call_api(
            url_filter,
            data=self.json_query,
            method="POST",
            params=querystring,
            method_whitelist=frozenset(
                ["HEAD", "TRACE", "GET", "PUT", "OPTIONS", "DELETE", "POST"]
            ),
        )
        self.total_hits = result
        return result

    def named(self, named_query, params=None, json_query=None):
        if json_query is not None and params is not None:
            raise Exception(
                "You can only specify either params or json_query, but not both!"
            )

        self.results = []
        self.named_query = named_query

        if json_query is not None:
            self.json_query = json_query
        else:
            if params is None:
                params = {}
            self.json_query = params

        return self

    def named_query_params(self, named_query):
        named = NamedQuery(self.carol)
        named.by_name(named_query=named_query)
        return named.param_dict

    def query(self, json_query):
        self.json_query = json_query
        return self

    def all(self, dm_name):
        if not dm_name.endswith("Golden"):
            dm_name = dm_name + "Golden"
        self.json_query = Filter.Builder().type(dm_name).build().to_json()
        self.index_type = "MASTER"
        return self

    def delete(self, json_query):
        # TODO: we should check the number of records to be deleted. If too many,
        # it can be a problem.
        self.json_query = json_query
        self.querystring = {"indexType": self.index_type}
        url_filter = "v2/queries/filter"
        result = self.carol.call_api(
            url_filter,
            data=self.json_query,
            params=self.querystring,
            method="DELETE",
            method_whitelist=frozenset(["HEAD", "TRACE", "GET", "PUT", "OPTIONS"]),
        )

        print("Deleted: ", result)


class ParQuery:
    def __init__(self, carol, backend="dask", return_df=True, verbose=50, n_jobs=4):
        """

        :param carol:
        :param backend:
        :param return_df:
        :param verbose:
        :param n_jobs:
        """
        import pandas as pd

        self._stag_mdm_key_range = None
        self._multiplier = None
        self.carol = carol
        self.return_df = return_df
        if return_df:
            import pandas as pd
        self.backend = backend
        self.verbose = verbose
        self.n_jobs = n_jobs

        assert self.backend == "dask" or self.backend == "joblib"

    def _get_min_max(
        self, datamodel_name, mdm_key, index_type, custom_filter=None, multiplier=None
    ):

        if custom_filter is not None:
            j = custom_filter
        else:
            j = (
                Filter.Builder()
                .type(datamodel_name)
                .aggregation_list(
                    [
                        MINIMUM(name="MINIMUM", params=mdm_key),
                        MAXIMUM(name="MAXIMUM", params=mdm_key),
                    ]
                )
                .build()
                .to_json()
            )

        query = (
            Query(
                self.carol,
                index_type=index_type,
                only_hits=False,
                get_aggs=True,
                save_results=False,
                print_status=True,
                page_size=1,
            )
            .query(j)
            .go()
        )

        if query.results[0].get("aggs") is None:
            return None, None, None
        sample = query.results[0].get("hits")[0]
        min_v = query.results[0]["aggs"]["MINIMUM"]["value"]
        max_v = query.results[0]["aggs"]["MAXIMUM"]["value"]
        print(f"Total Hits to download: {query.total_hits}")

        if multiplier is not None:
            min_v = int(min_v * multiplier) - 10
            max_v = int(max_v * multiplier) + 10
        return min_v, max_v, sample

    def _get_staging_from_golden_rejected(
        self, datamodel_name, connector_id, staging_name, fields
    ):

        index_type = "STAGING"
        self.datamodel_name = f"{datamodel_name}Rejected"
        self.fields = "mdmStagingRecord"
        self.filter_stag = f"{connector_id}_{staging_name}"
        only_hits = False
        mdm_key = "mdmStagingRecord.mdmCounterForEntity"
        j = (
            Filter.Builder()
            .type(self.datamodel_name)
            .must(TERM_FILTER(key="mdmStagingEntityName.raw", value=self.filter_stag))
            .aggregation_list(
                [
                    MINIMUM(name="MINIMUM", params=mdm_key),
                    MAXIMUM(name="MAXIMUM", params=mdm_key),
                ]
            )
            .build()
            .to_json()
        )

        min_v, max_v, sample = self._get_min_max(
            datamodel_name=self.datamodel_name,
            mdm_key=mdm_key,
            index_type=index_type,
            custom_filter=j,
        )
        if (min_v is None) and (max_v is None):
            return []
        chunks = ranges(min_v, max_v, self.slices)

        # rejected

        print(f"Number of chunks for rejected: {len(chunks)}")
        self.fields_to_get = [
            self.fields + "." + i
            for i in sample.get(self.fields).keys()
            for j in fields
            if j + "_" in i
        ]

        self.custom_filter = (
            Filter.Builder()
            .type(self.datamodel_name)
            .must(TERM_FILTER(key="mdmStagingEntityName.raw", value=self.filter_stag))
        )

        if self.backend == "dask":
            list_to_compute = _dask_backend(
                carol=self.carol,
                chunks=chunks,
                datamodel_name=self.datamodel_name,
                page_size=self.page_size,
                index_type=index_type,
                fields=self.fields,
                only_hits=only_hits,
                mdm_key=mdm_key,
                return_df=self.return_df,
                fields_to_get=self.fields_to_get,
                custom_filter=self.custom_filter,
            )
        elif self.backend == "joblib":
            list_to_compute = _joblib_backend(
                carol=self.carol,
                chunks=chunks,
                datamodel_name=self.datamodel_name,
                page_size=self.page_size,
                index_type=index_type,
                fields=self.fields,
                only_hits=only_hits,
                mdm_key=mdm_key,
                return_df=self.return_df,
                fields_to_get=self.fields_to_get,
                custom_filter=self.custom_filter,
                n_jobs=self.n_jobs,
                verbose=self.verbose,
            )
        else:
            raise KeyError

        return list_to_compute

    def _get_golden(self, datamodel_name=None, fields=None):

        index_type = "MASTER"
        self.datamodel_name = f"{datamodel_name}Golden"
        self.fields = "mdmGoldenFieldAndValues"
        self.fields_to_get = [
            self.fields + "." + i if self.fields not in i else i for i in fields
        ]
        only_hits = True
        mdm_key = "mdmCounterForEntity"

        min_v, max_v, sample = self._get_min_max(
            datamodel_name=self.datamodel_name, mdm_key=mdm_key, index_type=index_type
        )
        if (min_v is None) and (max_v is None):
            return []
        chunks = ranges(min_v, max_v, self.slices)
        print(f"Number of chunks: {len(chunks)}")

        if self.backend == "dask":
            list_to_compute = _dask_backend(
                carol=self.carol,
                chunks=chunks,
                datamodel_name=self.datamodel_name,
                page_size=self.page_size,
                index_type=index_type,
                fields=self.fields,
                only_hits=only_hits,
                mdm_key=mdm_key,
                return_df=self.return_df,
                fields_to_get=self.fields_to_get,
                custom_filter=self.custom_filter,
            )
        elif self.backend == "joblib":
            list_to_compute = _joblib_backend(
                carol=self.carol,
                chunks=chunks,
                datamodel_name=self.datamodel_name,
                page_size=self.page_size,
                index_type=index_type,
                fields=self.fields,
                only_hits=only_hits,
                mdm_key=mdm_key,
                return_df=self.return_df,
                fields_to_get=self.fields_to_get,
                custom_filter=self.custom_filter,
                n_jobs=self.n_jobs,
                verbose=self.verbose,
            )

        if self.return_df:
            return pd.concat(list_to_compute, ignore_index=True, sort=True)
        list_to_compute = list(itertools.chain(*list_to_compute))
        return list_to_compute

    def _get_staging(
        self, connector_id=None, connector_name=None, staging_name=None, fields=None
    ):

        if not connector_id:
            connector_id = Connectors(self.carol).get_by_name(connector_name)["mdmId"]

        index_type = "STAGING"
        self.datamodel_name = f"{connector_id}_{staging_name}"
        self.fields_to_get = fields
        self.fields = None
        only_hits = False
        mdm_key = "mdmCounterForEntity"
        if self._stag_mdm_key_range is not None:
            mdm_key = self._stag_mdm_key_range

        min_v, max_v, sample = self._get_min_max(
            datamodel_name=self.datamodel_name,
            mdm_key=mdm_key,
            index_type=index_type,
            multiplier=self._multiplier,
        )
        if (min_v is None) and (max_v is None):
            return []
        chunks = ranges(min_v, max_v, self.slices)
        print(f"Number of chunks: {len(chunks)}")

        if self.backend == "dask":
            list_to_compute = _dask_backend(
                carol=self.carol,
                chunks=chunks,
                datamodel_name=self.datamodel_name,
                page_size=self.page_size,
                index_type=index_type,
                fields=self.fields,
                only_hits=only_hits,
                mdm_key=mdm_key,
                return_df=self.return_df,
                fields_to_get=self.fields_to_get,
                custom_filter=self.custom_filter,
            )
        elif self.backend == "joblib":
            list_to_compute = _joblib_backend(
                carol=self.carol,
                chunks=chunks,
                datamodel_name=self.datamodel_name,
                page_size=self.page_size,
                index_type=index_type,
                fields=self.fields,
                only_hits=only_hits,
                mdm_key=mdm_key,
                return_df=self.return_df,
                fields_to_get=self.fields_to_get,
                custom_filter=self.custom_filter,
                n_jobs=self.n_jobs,
                verbose=self.verbose,
            )
        else:
            raise KeyError

        if self.return_df:
            return pd.concat(list_to_compute, ignore_index=True, sort=True)
        list_to_compute = list(itertools.chain(*list_to_compute))
        return list_to_compute

    def go(
        self,
        datamodel_name=None,
        slices=1000,
        page_size=1000,
        staging_name=None,
        connector_id=None,
        connector_name=None,
        fields=None,
    ):
        assert slices < 9999, "10k is the largest slice possible"
        self.slices = slices
        self.page_size = page_size
        if fields is None:
            fields = []
        self.custom_filter = None

        if datamodel_name is None:
            assert connector_id or connector_name
            assert staging_name

            return self._get_staging(
                connector_id=connector_id,
                connector_name=connector_name,
                staging_name=staging_name,
                fields=fields,
            )
        else:

            return self._get_golden(datamodel_name=datamodel_name, fields=fields)


def _dask_backend(
    carol,
    chunks,
    datamodel_name,
    page_size,
    index_type,
    fields,
    only_hits,
    mdm_key,
    return_df,
    fields_to_get,
    custom_filter,
):
    list_to_compute = []
    for RANGE_FILTER in chunks:
        y = dask.delayed(_par_query)(
            datamodel_name=datamodel_name,
            RANGE_FILTER=RANGE_FILTER,
            page_size=page_size,
            login=carol,
            index_type=index_type,
            fields=fields,
            only_hits=only_hits,
            mdm_key=mdm_key,
            return_df=return_df,
            fields_to_get=fields_to_get,
            custom_filter=custom_filter,
        )
        list_to_compute.append(y)

    return dask.compute(*list_to_compute)


def _joblib_backend(
    carol,
    chunks,
    datamodel_name,
    page_size,
    index_type,
    fields,
    only_hits,
    mdm_key,
    return_df,
    fields_to_get,
    custom_filter,
    n_jobs,
    verbose,
):
    from joblib import Parallel, delayed

    list_to_compute = Parallel(n_jobs=n_jobs, verbose=verbose)(
        delayed(_par_query)(
            datamodel_name=datamodel_name,
            RANGE_FILTER=RANGE_FILTER,
            page_size=page_size,
            login=carol,
            index_type=index_type,
            fields=fields,
            only_hits=only_hits,
            mdm_key=mdm_key,
            return_df=return_df,
            fields_to_get=fields_to_get,
            custom_filter=custom_filter,
        )
        for RANGE_FILTER in chunks
    )
    return list_to_compute


def _par_query(
    datamodel_name,
    RANGE_FILTER,
    page_size=1000,
    login=None,
    index_type="MASTER",
    fields=None,
    mdm_key=None,
    only_hits=True,
    return_df=True,
    fields_to_get=None,
    custom_filter=None,
):
    if custom_filter is not None:
        json_query = copy.deepcopy(custom_filter)
        json_query = (
            json_query.must(RF(key=mdm_key, value=RANGE_FILTER)).build().to_json()
        )

    else:
        json_query = (
            Filter.Builder()
            .type(datamodel_name)
            .must(RF(key=mdm_key, value=RANGE_FILTER))
            .build()
            .to_json()
        )

    query = (
        Query(
            login,
            page_size=page_size,
            save_results=False,
            print_status=False,
            index_type=index_type,
            only_hits=only_hits,
            fields=fields_to_get,
        )
        .query(json_query)
        .go()
    )
    query = query.results

    if not only_hits:
        query = [i["hits"] for i in query]
        query = list(itertools.chain(*query))
        if fields:
            query = [elem.get(fields, elem) for elem in query if elem.get(fields, None)]

    if return_df:
        return pd.DataFrame(query)

    return query


def _write_results(filepath: str, results) -> None:
    with open(filepath, "a", encoding="utf8") as file:
        file.write(json.dumps(results, ensure_ascii=False))
        file.write("\n")



"""Carol app funtionalities."""
import io
from pathlib import Path
import typing as T
import zipfile

import requests

from . import exceptions
from .carol import Carol
from .utils.deprecation_msgs import _deprecation_msgs
from .utils.miscellaneous import zip_folder


class Apps:

    """Carol App instance.

    Args:
        carol: Carol() instance
    """

    def __init__(self, carol: Carol):
        self.carol = carol

        self.all_apps: T.Optional[T.Dict] = None
        self.app_settings: T.Optional[T.Dict] = None
        self.current_app: T.Optional[T.Dict] = None
        self.current_app_id: T.Optional[str] = None
        self.current_app_name: T.Optional[str] = None
        self.full_settings: T.Optional[T.Dict] = None
        self.total_hits: T.Optional[int] = None

    def _build_query_params(
        self,
        entity_space="PRODUCTION",
        offset=0,
        page_size=100,
        sort_by=None,
        sort_order=None,
    ) -> T.Dict[str, T.Any]:
        if sort_by is None:
            return {
                "offset": offset,
                "pageSize": page_size,
                "entitySpace": entity_space,
            }
        return {
            "offset": offset,
            "pageSize": page_size,
            "sortOrder": sort_order,
            "sortBy": sort_by,
            "entitySpace": entity_space,
        }

    def _define_current_run(self, query: T.Dict) -> None:
        self.current_app_id = query.get("mdmId")
        self.current_app_name = query.get("mdmName")
        self.current_app = {self.current_app_name: query}

    def all(
        self,
        entity_space: str = "PRODUCTION",
        page_size: int = 50,
        offset: int = 0,
        sort_order: str = "ASC",
        sort_by: T.Optional[str] = None,
    ) -> T.Dict:
        """Get all app information in this environment.

        Args:
            entity_space: Space to get the app from. Possible values:

                    1. PRODUCTION: For production.
                    2. WORKING: For Draft apps.

            offset: Offset for pagination. Only used when `scrollable=False`.
            page_size: Number of records downloaded in each pagination. The maximum
                value is 1000.
            sort_order: Sort ascending ('ASC') vs. descending ('DESC').
            sort_by: Name to sort by.

        Returns:
            All apps json definition

        Raises:
            NotMapAsCallResponseException if call_api return is not a dict.
        """
        query_string = self._build_query_params(
            offset=offset,
            page_size=page_size,
            entity_space=entity_space,
            sort_by=sort_by,
            sort_order=sort_order,
        )
        query = self.carol.call_api("v1/tenantApps", method="GET", params=query_string)
        if not isinstance(query, dict):
            raise exceptions.NotMapAsCallResponseException

        self.all_apps = {app["mdmName"]: app for app in query["hits"]}
        self.total_hits = query["totalHits"]
        return self.all_apps

    def get_by_name(self, app_name: str, entity_space: str = "PRODUCTION") -> T.Dict:
        """Get app information by app name.

        Args:
            app_name: App name.
            entity_space: Space to get the app information from. Possible values

                    1. PRODUCTION: For production.
                    2. WORKING: For Draft apps.

        Returns:
            App info json definition.

        Raises:
            NotMapAsCallResponseException if call_api return is not a dict.
        """
        query_string = {"entitySpace": entity_space}
        url = f"v1/tenantApps/name/{app_name}"
        query = self.carol.call_api(url, method="GET", params=query_string)
        if not isinstance(query, dict):
            raise exceptions.NotMapAsCallResponseException

        self._define_current_run(query)

        return query

    def get_by_id(self, app_id: str, entity_space: str = "PRODUCTION") -> T.Dict:
        """Get app information by ID.

        Args:
            app_id: App id.
            entity_space: Space to get the app information from. Possible values

                    1. PRODUCTION: For production.
                    2. WORKING: For Draft apps.

        Returns:
            App info json definition.

        Raises:
            NotMapAsCallResponseException if call_api return is not a dict.
        """
        query_string = {"entitySpace": entity_space}
        url = f"v1/tenantApps/{app_id}"
        query = self.carol.call_api(url, method="GET", params=query_string)
        if not isinstance(query, dict):
            raise exceptions.NotMapAsCallResponseException

        self._define_current_run(query)

        return query

    def get_settings(
        self,
        app_name: T.Optional[str] = None,
        app_id: T.Optional[str] = None,
        entity_space: str = "PRODUCTION",
        check_all_spaces: bool = False,
    ) -> T.Dict:
        """Get settings from app.

        Settings are the settings available in Carol's UI.

        Args:
            app_name: App name, if None will get from `app_id`.
            app_id: App id. Either app_name or app_id must be set.
            entity_space: Space to get the app settings from. Possible values

                    1. PRODUCTION: For production.
                    2. WORKING: For Draft apps.

            check_all_spaces: Check all entity spaces.

        Returns:
            Settings.

        Raises:
            Exception if carol.app_name or parameter app_name are both None.
            NotMapAsCallResponseException if call_api return is not a dict.
        """
        assert app_name or app_id or self.carol.app_name

        if app_id is not None:
            self.get_by_id(app_id, entity_space)
        elif app_name is not None:
            self.get_by_name(app_name, entity_space)
        elif self.carol.app_name is not None:
            self.get_by_name(self.carol.app_name, entity_space)
        else:
            raise Exception("carol.app_name is None.")

        query_string = {"entitySpace": entity_space, "checkAllSpaces": check_all_spaces}

        query = self.carol.call_api(
            f"v1/tenantApps/{self.current_app_id}/settings",
            method="GET",
            params=query_string,
        )
        if not isinstance(query, dict):
            raise exceptions.NotMapAsCallResponseException

        self.app_settings = {}
        self.full_settings = {}

        if not isinstance(query, list):
            query = [query]

        for query_el in query:
            self.app_settings.update(
                {
                    i["mdmName"]: i.get("mdmParameterValue")
                    for i in query_el.get("mdmTenantAppSettingValues", {})
                }
            )
            self.full_settings.update(
                {i["mdmName"]: i for i in query_el.get("mdmTenantAppSettingValues", {})}
            )

        return self.app_settings

    def download_app(
        self,
        app_version: str,
        app_name: T.Optional[str] = None,
        carolappname: T.Optional[str] = None,
        carolappversion: T.Optional[str] = None,
        file_path: str = "carol.zip",
        extract: bool = False,
    ) -> None:
        """Download App artifacts.

        Args:
            app_name: Carol app name. It will overwrite the app name used in Carol()
                initialization.
            app_version: App Version
            carolappname: App Name. Deprecated. Use app_name
            carolappversion: App Version. Deprecated. Use app_version
            file_path: Path to save the zip file.
            extract: Either extract the zip files or not.

        Raises:
            ValueError when app_version is not set.
        """
        if carolappname is not None:
            app_name = carolappname
            _deprecation_msgs("`carolappname` is deprecated use `app_name`.")
        if carolappversion is not None:
            app_version = carolappversion
            _deprecation_msgs("`carolappversion` is deprecated use `app_version`.")

        if app_name is None:
            app_name = self.carol.app_name

        if app_version is None:
            raise ValueError("app_version must be set.")

        url = f"v1/carolApps/download/{app_name}/version/{app_version}"

        response = self.carol.call_api(
            url, method="GET", stream=True, downloadable=True
        )
        if not isinstance(response, requests.Response):
            raise exceptions.NotResponseAsCallResponseException

        if extract is True:
            with zipfile.ZipFile(io.BytesIO(response.content)) as zip_file:
                zip_file.extractall(file_path)
        else:
            with open(file_path, "wb") as out:  # Open temporary file as bytes
                out.write(io.BytesIO(response.content).read())

    def get_manifest(self, app_name: T.Optional[str] = None) -> T.Dict:
        """Get manifest file.

        Args:
            app_name: Carol app name. It will overwrite the app name used in Carol()
                initialization.

        Returns:
            Manifest file.

        Raises:
            NotMapAsCallResponseException if call_api return is not a dict.
        """
        if app_name is None:
            app_name = self.carol.app_name

        url = f"v1/tenantApps/manifest/{app_name}"

        response = self.carol.call_api(url, method="GET")
        if not isinstance(response, dict):
            raise exceptions.NotMapAsCallResponseException

        return response

    def edit_manifest(
        self, manifest: T.Dict, app_name: T.Optional[str] = None
    ) -> T.Dict:
        """Edit manifest file.

        Args:
            manifest: Dictionary with the manifest
            app_name: Carol app name. It will overwrite the app name used in Carol()
                initialization.

        Returns:
            Carol API response.

        Raises:
            NotMapAsCallResponseException if call_api return is not a dict.
        """
        if app_name is None:
            app_name = self.carol.app_name

        url = f"v1/tenantApps/manifest/{app_name}"

        response = self.carol.call_api(url, method="PUT", data=manifest)
        if not isinstance(response, dict):
            raise exceptions.NotMapAsCallResponseException

        return response

    def get_git_process(self, app_name: T.Optional[str] = None) -> T.List[T.Dict]:
        """Get Git processes defined in the manifest file.

        Args:
            app_name: Carol app name. It will overwrite the app name used in Carol()
                initialization.

        Returns:
            List of Dicts.

        Raises:
            NotListAsCallResponseException if call_api return is not a list.
        """
        if app_name is None:
            app_name = self.carol.app_name

        url = f"v1/compute/{app_name}/getProcessesByGitRepo"
        response = self.carol.call_api(url, method="POST")
        if not isinstance(response, list):
            raise exceptions.NotListAsCallResponseException

        return response

    def build_docker_git(
        self,
        git_token: str,
        app_name: T.Optional[str] = None,
    ) -> T.List:
        """Build all images listed in the manifest definition.

        Args:
            git_token: Git token to be used to pull the files.
            app_name: Carol app name. It will overwrite the app name used in Carol()
                initialization.

        Returns:
            List of task ids.
        """
        if app_name is None:
            app_name = self.carol.app_name

        manifest = self.get_git_process(app_name)

        self._assert_manifest_fields(manifest)

        tasks = []
        url = f"v1/compute/{app_name}/buildGitDocker"
        for build in manifest:
            docker_name = build["dockerName"]
            docker_tag = build["dockerTag"]
            instance_type = build["instanceType"]
            params = {
                "dockerName": docker_name,
                "tagName": docker_tag,
                "gitToken": git_token,
                "instanceType": instance_type,
            }
            response = self.carol.call_api(url, method="POST", params=params)

            tasks.append(response)

        return tasks

    @staticmethod
    def _assert_manifest_fields(manifest: T.List[T.Dict]) -> None:
        """Assert that the fields needed to build the image exist.

        Args:
            manifest: list of docker definition in the manifest file.

        Raises:
            ValueError when docker definition is missing.

        Raises:
            ValueError when docker definition is missing.
        """
        fields = {"dockerName", "dockerTag", "instanceType"}
        for build in manifest:
            set_diff = fields - set(build)
            if len(set_diff) >= 1:
                raise ValueError(f"Missing docker definition {set_diff}")

    def update_setting_values(
        self, settings: T.Dict, app_name: T.Optional[str] = None
    ) -> T.Dict:
        """Change Settings values in Carol.

        Args:
            settings: dict with settings: {"param1": "value1", "param2": "value2"}
            app_name: App name to change the settings.

        Returns:
            Carol Response.

        Raises:
            ValueError if settings type is not a dictionary.
            NotMapAsCallResponseException if call_api return is not a dict.
        """
        app_name = app_name or self.carol.app_name
        if app_name is None:
            raise Exception("carol.app_name is None.")

        if not isinstance(settings, dict):
            ValueError("settings should be a dictionary.")

        app_id = self.get_by_name(app_name)["mdmId"]
        settings_id = self._get_app_settings_config(app_id=app_id)["mdmId"]
        data = [{"mdmName": i, "mdmParameterValue": j} for i, j in settings.items()]

        url = f"v1/tenantApps/{app_id}/settings/{settings_id}?publish=true"
        response = self.carol.call_api(url, method="PUT", data=data)
        if not isinstance(response, dict):
            raise exceptions.NotMapAsCallResponseException

        return response

    def _get_app_settings_config(
        self, app_name: T.Optional[str] = None, app_id: T.Optional[str] = None
    ) -> T.Dict:
        if app_id is None:
            app_name = app_name or self.carol.app_name
            if app_name is None:
                raise Exception("carol.app_name is None.")

            app_id = self.get_by_name(app_name)["mdmId"]

        response = self.carol.call_api(path=f"v1/tenantApps/{app_id}/settings")
        if not isinstance(response, dict):
            raise exceptions.NotMapAsCallResponseException

        return response

    def start_app_process(
        self,
        process_name: str,
        app_name: T.Optional[str] = None,
    ) -> T.Dict:
        """Start a carol process by process name.

        Args:
            process_name: Process name.
            app_name: App name to change the settings.

        Returns:
            task information.

        Raises:
            Exception if carol.app_name or parameter app_name are both None.
            NotMapAsCallResponseException if call_api return is not a dict.
        """
        app_name = app_name or self.carol.app_name
        if app_name is None:
            raise Exception("carol.app_name is None.")

        app_id = self.get_by_name(app_name)["mdmId"]
        process_id = self.get_processes_info(app_name=app_name)["mdmId"]
        url = f"v1/tenantApps/{app_id}/aiprocesses/{process_id}/execute/{process_name}"
        response = self.carol.call_api(url, method="POST")
        if not isinstance(response, dict):
            raise exceptions.NotMapAsCallResponseException

        return response

    def get_processes_info(
        self,
        app_name: T.Optional[str] = None,
        entity_space: str = "WORKING",
        check_all_spaces: bool = True,
    ) -> T.Dict:
        """Get app processes information.

        Args:
            app_name: App name to change the settings.
            entity_space: WORKING or PRODUCTION
            check_all_spaces: Check all spaces.

        Returns:
            Process informations

        Raises:
            Exception if carol.app_name or parameter app_name are both None.
            NotMapAsCallResponseException if call_api return is not a dict.
        """
        params = {"entitySpace": entity_space, "checkAllSpaces": check_all_spaces}
        app_name = app_name or self.carol.app_name
        if app_name is None:
            raise Exception("carol.app_name is None.")
        app_id = self.get_by_name(app_name)["mdmId"]
        url = f"v1/tenantApps/{app_id}/aiprocesses"
        response = self.carol.call_api(url, method="GET", params=params)
        if not isinstance(response, dict):
            raise exceptions.NotMapAsCallResponseException

        return response

    def get_subscribable_carol_apps(self) -> T.List:
        """Find all available apps to install in this env.

        Returns:
            list of apps.

        Raises:
            NotMapAsCallResponseException if call_api return is not a dict.
        """
        url = "v1/tenantApps/subscribableCarolApps"
        response = self.carol.call_api(url, method="GET")
        if not isinstance(response, dict):
            raise exceptions.NotMapAsCallResponseException

        return response["hits"]

    def install_carol_app(
        self,
        app_name: str = None,
        app_version: str = None,
        connector_group: str = None,
        publish: bool = True,
    ) -> T.Optional[T.Dict]:
        """Install a carol app in an env.

        Args:
            app_name:  App name to change the settings.
            app_version: App version to install. If not specified, it will install the
                most recent.
            connector_group: Connector Group to install.
            publish: If publish the update.

        Returns:
            Carol task.

        Raises:
            NotMapAsCallResponseException if call_api return is not a dict.
        """
        if app_name is None:
            app_name = self.carol.app_name

        to_install = self.get_subscribable_carol_apps()
        to_install = [i for i in to_install if i["mdmName"] == app_name]

        if app_version is None:
            to_install = sorted(to_install, key=lambda x: x["mdmAppVersion"])
        else:
            to_install = [i for i in to_install if i["mdmAppVersion"] == app_version]

        if len(to_install) > 0:
            to_install_first = to_install[0]
        else:
            return None

        to_install_id = to_install_first["mdmId"]

        url = f"v1/tenantApps/subscribe/carolApps/{to_install_id}"
        updated = self.carol.call_api(url, method="POST")
        if not isinstance(updated, dict):
            raise exceptions.NotMapAsCallResponseException

        params = {"publish": publish, "connectorGroup": connector_group}
        url = f"v1/tenantApps/{updated['mdmId']}/install"
        response = self.carol.call_api(url, method="POST", params=params)
        if not isinstance(response, dict):
            raise exceptions.NotMapAsCallResponseException
        return response

    def get_app_details(
        self, app_name: str = None, entity_space: str = "PRODUCTION"
    ) -> T.Dict:
        """Find all information about an app.

        This will fetch information about connector groups, AI process, descriptions,
        data models, etc.

        Args:
            app_name: App name to change the settings.
            entity_space: WORKING or PRODUCTION

        Returns:
            Carol response.

        Raises:
            Exception if carol.app_name or parameter app_name are both None.
            NotMapAsCallResponseException if call_api return is not a dict.
        """
        app_name = app_name or self.carol.app_name
        if app_name is None:
            raise Exception("carol.app_name is None.")

        # check if it exists as a subscribable apps
        to_install = self.get_subscribable_carol_apps()
        to_install = [i for i in to_install if i["mdmName"] == app_name]

        if len(to_install) > 0:
            to_install_first = to_install[0]
            app_id = to_install_first["mdmId"]
        else:
            # try installed app.
            app_id = self.get_by_name(app_name)["mdmCarolAppId"]

        params = {"entitySpace": entity_space}
        url = f"v1/carolApps/{app_id}/details"
        response = self.carol.call_api(url, method="GET", params=params)
        if not isinstance(response, dict):
            raise exceptions.NotMapAsCallResponseException

        return response

    def upload_file(
        self,
        filepath: str,
        app_name: str = None,
    ) -> T.Dict:
        """Upload the app artifacts.

        Args:
            filepath: Folder path with artifacts (manifest.json or site)
            app_name: App name

        Returns:
            Carol response.

        Raises:
            Exception if carol.app_name or parameter app_name are both None.
            NotMapAsCallResponseException if call_api return is not a dict.
        """
        app_name = app_name or self.carol.app_name
        if app_name is None:
            raise Exception("carol.app_name is None.")

        app_id = self.get_by_name(app_name)["mdmCarolAppId"]

        if Path(filepath).is_dir():
            filepath = zip_folder(filepath)

        if not filepath.endswith((".zip", ".json")):
            raise Exception("File must be a .zip or .json.")

        with open(filepath, "rb") as file_:
            files = {"file": file_}

        url = f"v1/carolApps/{app_id}/files/upload"
        response = self.carol.call_api(
            url, method="POST", files=files, content_type=None
        )
        if not isinstance(response, dict):
            raise exceptions.NotMapAsCallResponseException

        return response



import functools
import typing as T
import warnings

import deprecation

from .. import __version__


def _deprecation_msgs(msg):
    warnings.warn(msg, DeprecationWarning, stacklevel=3)


def deprecation_msg(deprecated_in: str, removed_in: str, details: str) -> str:
    """Create a message for deprecation.

    Args:
        deprecated_in: version of deprecation.
        removed_in: version of removal.
        details: deprecation message.

    Returns:
        deprecation message.
    """
    if deprecated_in >= __version__:
        msg = f"This function will be deprecated at version {deprecated_in}. "
    else:
        msg = f"This function was deprecated at version {deprecated_in}. "

    msg += f"It will be removed at {removed_in}. "
    msg += details
    return msg


def deprecated(deprecated_in, removed_in, details) -> T.Callable:
    """Decorate deprecated functions.

    Args:
        deprecated_in: version of deprecation.
        removed_in: version of removal.
        details: deprecation message.
    """
    def _outer_func(original_func) -> T.Callable:
        func_ = deprecation.deprecated(deprecated_in, removed_in, __version__, details)
        deprecated_func = func_(original_func)

        @functools.wraps(deprecated_func)
        def _inner_func(*args, **kwargs) -> T.Any:
            msg = deprecation_msg(deprecated_in, removed_in, details)
            warnings.warn(msg)
            return deprecated_func(*args, **kwargs)

        return _inner_func

    return _outer_func



from datetime import datetime, timedelta
import json
from pathlib import Path
from unittest import mock
import typing as T

import pycarol


def test_token_init() -> None:
    """Test the initialization of the Token class in the pycarol.bigquery module."""
    token_mock = mock.MagicMock()
    service_account = {"expiration_time": "expiration_time"}
    env: T.Dict = {}
    pycarol.bigquery.Token.__init__(token_mock, service_account, env)
    assert token_mock._env == env
    assert token_mock.expiration_time == service_account["expiration_time"]
    assert token_mock.service_account == service_account


def test_token_to_dict() -> None:
    """Test the to_dict() method of the Token class in the pycarol.bigquery module."""
    token_mock = mock.MagicMock()
    token_mock.service_account = {}
    token_mock._env = {}
    token_mock.expiration_time = ""
    _dict = pycarol.bigquery.Token.to_dict(token_mock)
    assert _dict == {"service_account": {}, "env": {}, "expiration_time": ""}


def test_token_expired() -> None:
    """Test the expired() method of the Token class in the pycarol.bigquery module."""
    token_mock = mock.MagicMock()
    dt_format = "%Y-%m-%dT%H:%M:%S.%fZ"
    pass_date = datetime.utcnow() - timedelta(1)
    token_mock.expiration_time = datetime.strftime(pass_date, dt_format)
    expired = pycarol.bigquery.Token.expired(token_mock)
    assert expired is True


@mock.patch("pycarol.bigquery.Storage")
def test_token_manager_init(storage_mock) -> None:
    """Test the initialization of the pycarol.bigquery.TokenManager class."""
    manager_mock = mock.MagicMock()
    carol_mock = mock.MagicMock()
    carol_mock.get_current.return_value = {"env_id": 5}
    pycarol.bigquery.TokenManager.__init__(manager_mock, carol_mock)
    assert manager_mock._carol == carol_mock
    assert manager_mock.expiration_window == 24
    assert manager_mock._env == {"env_id": 5}
    base_dir = Path(pycarol.__TEMP_STORAGE__)
    assert manager_mock._tmp_filepath == base_dir / ".pycarol_temp_5.json"
    assert manager_mock.cache_cds is True
    assert manager_mock._storage == storage_mock.return_value
    assert manager_mock.token == manager_mock.get_token.return_value


def test_token_manager_issue_new_key() -> None:
    """Test the _issue_new_key() method of the pycarol.bigquery.TokenManager class."""
    manager_mock = mock.MagicMock()
    manager_mock._carol = mock.MagicMock()
    manager_mock._carol.call_api.return_value = {}
    key = pycarol.bigquery.TokenManager._issue_new_key(manager_mock)
    assert key == {}


def test_token_manager_save_token_file() -> None:
    """Test the _save_token_file() method of the pycarol.bigquery.TokenManager class."""
    manager_mock = mock.MagicMock()
    manager_mock._tmp_filepath = Path("/tmp/pycarol_test/test_sa.env")
    token_mock = mock.MagicMock()
    token_mock.to_dict.return_value = {"test": "test"}
    pycarol.bigquery.TokenManager._save_token_file(manager_mock, token_mock)
    with open(manager_mock._tmp_filepath, "r", encoding="utf-8") as file:
        sa = json.load(file)
    assert sa == {"test": "test"}

    Path("/tmp/pycarol_test/test_sa.env").unlink()


def test_token_manager_save_token_cloud() -> None:
    """Test the _save_token_cloud() method of the pycarol.bigquery.TokenManager."""
    manager_mock = mock.MagicMock()
    Path("/tmp/pycarol_test/test_sa.env").touch()
    pycarol.bigquery.TokenManager._save_token_cloud(manager_mock)
    assert manager_mock._storage.save.called is True
    Path("/tmp/pycarol_test/test_sa.env").unlink()


@mock.patch("pycarol.bigquery.Token")
def test_token_manager_load_token_file(token_mock) -> None:
    """Test the _load_token_file() method of the pycarol.bigquery.TokenManager class."""
    sa = {"service_account": "test", "env": "test"}
    test_path = Path("/tmp/pycarol_test/test_sa.env")
    with open(test_path, "w", encoding="utf-8") as file:
        json.dump(sa, file)

    manager_mock = mock.MagicMock()
    manager_mock._tmp_filepath = test_path
    token = pycarol.bigquery.TokenManager._load_token_file(manager_mock)
    assert token == token_mock.return_value
    test_path.unlink()


def test_token_manager_load_token_cloud() -> None:
    """Test the _load_token_cloud() method of the pycarol.bigquery.TokenManager."""
    manager_mock = mock.MagicMock()
    test_path = Path("/tmp/pycarol_test/test_sa.env")
    test_path.touch()
    manager_mock._storage.exists.return_value = True
    manager_mock._storage.load.return_value = "/tmp/pycarol_test/test_sa.env"
    manager_mock._tmp_filepath = Path("/tmp/pycarol_test/test_sa2.env")
    token = pycarol.bigquery.TokenManager._load_token_cloud(manager_mock)
    assert token == manager_mock._load_token_file.return_value


@mock.patch("pycarol.bigquery.Token")
def test_token_manager_get_forced_token(token_mock) -> None:
    manager_mock = mock.MagicMock()
    token = pycarol.bigquery.TokenManager.get_forced_token(manager_mock)
    assert token == token_mock.return_value


@mock.patch("pycarol.bigquery.Token")
def test_token_manager_get_token(token_mock) -> None:
    """Test the get_token() method of the pycarol.bigquery.TokenManager class."""
    manager_mock = mock.MagicMock()
    token = pycarol.bigquery.TokenManager.get_token(manager_mock)
    assert token == manager_mock.get_forced_token.return_value


@mock.patch("pycarol.bigquery.TokenManager")
def test_bq_init(manager_mock) -> None:
    """Test the initialization of the BQ class in the pycarol.bigquery module."""
    bq_mock = mock.MagicMock()
    carol_mock = mock.MagicMock()
    carol_mock.get_current.return_value = {"env_id": "5"}
    pycarol.bigquery.BQ.__init__(bq_mock, carol_mock)
    assert bq_mock._env == {"env_id": "5"}
    assert bq_mock._project_id == "carol-5"
    assert bq_mock._dataset_id == "carol-5.5"
    assert bq_mock._token_manager == manager_mock.return_value


@mock.patch("pycarol.bigquery.Credentials")
@mock.patch("pycarol.bigquery.bigquery")
def test_bq_generate_client(bigquery_mock, credentials_mock) -> None:
    """Test the _generate_client() method of the pycarol.bigquery.BQ class."""
    sa = {"project_id": ""}
    client = pycarol.bigquery.BQ._generate_client(sa)
    assert client == bigquery_mock.Client.return_value


@mock.patch("pycarol.bigquery.bigquery.QueryJobConfig")
def test_bq_query_pd(query_job_mock) -> None:
    """Test the query() method of the pycarol.bigquery.BQ class."""
    import pandas as pd

    bq_mock = mock.MagicMock()
    query_ret = [
        {"col1": "val1", "col2": "val2"},
        {"col1": "val1", "col2": "val2"},
    ]
    client_mock = mock.MagicMock()
    client_mock.query.return_value = query_ret
    bq_mock._generate_client.return_value = client_mock
    query = ""
    ret = pycarol.bigquery.BQ.query(bq_mock, query, return_dataframe=True)
    assert ret.equals(pd.DataFrame(query_ret))  # type: ignore


@mock.patch("pycarol.bigquery.bigquery.QueryJobConfig")
def test_bq_query(query_job_mock) -> None:
    """Test the query() method of the pycarol.bigquery.BQ class."""
    bq_mock = mock.MagicMock()
    query_ret = [
        {"col1": "val1", "col2": "val2"},
        {"col1": "val1", "col2": "val2"},
    ]
    client_mock = mock.MagicMock()
    client_mock.query.return_value = query_ret
    bq_mock._generate_client.return_value = client_mock
    query = ""
    ret = pycarol.bigquery.BQ.query(bq_mock, query, return_dataframe=False)
    assert ret == query_ret


@mock.patch("pycarol.bigquery.TokenManager")
def test_storage_init(manager_mock) -> None:
    """Test the initialization of the BQStorage class in the pycarol.bigquery module."""
    storage_mock = mock.MagicMock()
    carol_mock = mock.MagicMock()
    carol_mock.get_current.return_value = {"env_id": "5"}
    pycarol.bigquery.BQStorage.__init__(storage_mock, carol_mock)
    assert storage_mock._env == {"env_id": "5"}
    assert storage_mock._project_id == "carol-5"
    assert storage_mock._dataset_id == "5"
    assert storage_mock._token_manager == manager_mock.return_value


@mock.patch("pycarol.bigquery.Credentials")
@mock.patch("pycarol.bigquery.bigquery_storage")
def test_storage_generate_client(bigquery_mock, credentials_mock) -> None:
    """Test the _generate_client() method of the pycarol.bigquery.BQStorage class."""
    sa = {"project_id": ""}
    client = pycarol.bigquery.BQStorage._generate_client(sa)
    assert client == bigquery_mock.BigQueryReadClient.return_value


@mock.patch("pycarol.bigquery.types")
def test_storage_get_read_session(types_mock) -> None:
    """Test the _get_read_session() method of the pycarol.bigquery.BQStorage class."""
    storage_mock = mock.MagicMock()
    client_mock = mock.MagicMock()
    session = pycarol.bigquery.BQStorage._get_read_session(
        storage_mock, client_mock, "table"
    )
    assert session == client_mock.create_read_session.return_value


def test_storage_query_pd() -> None:
    """Test the query() method of the pycarol.bigquery.BQStorage class."""
    import pandas as pd

    pages = mock.MagicMock()
    page1 = mock.MagicMock()
    page1.to_dataframe.return_value = pd.DataFrame(
        [
            {"col1": "name1", "col2": "name2"},
            {"col1": "name1", "col2": "name2"},
        ]
    )
    page2 = mock.MagicMock()
    page2.to_dataframe.return_value = pd.DataFrame(
        [
            {"col1": "name1", "col2": "name2"},
            {"col1": "name1", "col2": "name2"},
        ]
    )
    pages.pages = [page1, page2]
    reader_mock = mock.MagicMock()
    reader_mock.rows.return_value = pages

    client_mock = mock.MagicMock()
    client_mock.read_rows.return_value = reader_mock

    storage_mock = mock.MagicMock()
    storage_mock._generate_client.return_value = client_mock
    ret = pycarol.bigquery.BQStorage.query(storage_mock, "table")

    ret_expected = pd.DataFrame(
        [
            {"col1": "name1", "col2": "name2"},
            {"col1": "name1", "col2": "name2"},
            {"col1": "name1", "col2": "name2"},
            {"col1": "name1", "col2": "name2"},
        ]
    )
    assert ret_expected.equals(ret)


def test_storage_query() -> None:
    """Test the query() method of the pycarol.bigquery.BQStorage class."""
    pages = mock.MagicMock()
    page1 = [
        {"col1": "name1", "col2": "name2"},
        {"col1": "name1", "col2": "name2"},
    ]
    page2 = [
        {"col1": "name1", "col2": "name2"},
        {"col1": "name1", "col2": "name2"},
    ]
    pages.pages = [page1, page2]
    reader_mock = mock.MagicMock()
    reader_mock.rows.return_value = pages

    client_mock = mock.MagicMock()
    client_mock.read_rows.return_value = reader_mock

    storage_mock = mock.MagicMock()
    storage_mock._generate_client.return_value = client_mock
    ret = pycarol.bigquery.BQStorage.query(
        storage_mock, "table", return_dataframe=False
    )

    ret_expected = [
        [
            {"col1": "name1", "col2": "name2"},
            {"col1": "name1", "col2": "name2"},
        ],
        [
            {"col1": "name1", "col2": "name2"},
            {"col1": "name1", "col2": "name2"},
        ],
    ]
    assert ret_expected == ret
